/**
 * Copyright (c) 2005, David Holm
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the project nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdarg.h>
#include "vsprintf.h"
#include "of.h"

static struct {
	prom_entry	entry;

	ihandle		stdin;
	ihandle		stdout;
} of_s;

void of_puts(char *s);

/**
 * Internal function for making procedure calls to Open Firmware
 * Parameters:
 *   service - Name of service wanted
 *   nargs   - Number of arguments
 *   nret    - Number of return values
 *   ...     - Arguments
 * Returns: Pointer to returned values (if any)
 */
static void *of_call(const char *service, int nargs, int nret, ...)
{
	va_list list;
	int     i;
	struct {
		const char *service;
		int         nargs;
		int         nret;
		void       *args[10];
	} of_args;

	of_args.service = service;
	of_args.nargs   = nargs;
	of_args.nret    = nret;

	va_start(list, nret);
	for (i = 0; i < nargs; ++i) {
		of_args.args[i] = va_arg(list, void *);
	}
	va_end(list);
	for (i = 0; i < nret; ++i) {
		of_args.args[i + nargs] = 0;
	}
	of_s.entry(&of_args);
	if (nret > 0) {
		return of_args.args[nargs];
	} else {
		return 0;
	}
}

void of_init(prom_entry entry)
{
	phandle chosen;

	of_s.entry = entry;

	chosen = of_find_device("/chosen");
	if (chosen == 0L) {
		of_exit();
	}

	of_get_property(chosen, "stdout", &of_s.stdout, sizeof(ihandle));
	of_get_property(chosen, "stdin",  &of_s.stdin,  sizeof(ihandle));
}

phandle of_find_device(const char *name)
{
	return of_call("finddevice", 1, 1, name);
}

int of_get_property(phandle package, const char *name, void *mem, unsigned int len)
{
	return (int) of_call("getprop", 4, 1, package, name, mem, len);
}

int of_exit()
{
	return (int) of_call("exit", 0, 1);
}

void of_puts(char *s)
{
	const char *p;
	const char *q;

	for (p = s; *p != 0; p = q) {
		for (q = p; *q != 0 && *q != '\n'; ++q);
		if (q > p) {
			of_call("write", 3, 1, of_s.stdout, p, q - p);
		}
		if (*q != 0) {
			++q;
			of_call("write", 3, 1, of_s.stdout, "\r\n", 2);
		}
	}
}

void of_vfprintf(char *fmt, va_list ap)
{
	static char printf_buf[2048];
	vsprintf(printf_buf, fmt, ap);
	of_puts(printf_buf);
}

void of_printf(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	of_vfprintf(fmt, ap);
	va_end(ap);
}

void *of_claim(void *vaddr, unsigned int len, unsigned int align)
{
	return (void *) of_call("claim", 3, 1, vaddr, len, align);
}

void of_release(void *vaddr, unsigned int len)
{
	of_call("release", 2, 0, vaddr, len);
}
